#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Member, GetAttr, SetAttr, DelAttr, reset_property


class Property(Member):
    """ A Member which behaves similar to a Python property.

    """
    __slots__ = ()

    def __init__(self, fget=None, fset=None, fdel=None, cached=False):
        """ Initialize a Property member.

        Parameters
        ----------
        fget : callable or None, optional
            The callable invoked to get the property value. It must
            accept a single argument which is the owner object. If not
            provided, the property cannot be read. The default is None.

        fset : callable or None, optional
            The callable invoked to set the property value. It must
            accept two arguments: the owner object and property value.
            If not provided, the property cannot be set. The default
            is None.

        fdel : callable or None, optional
            The callable invoked to delete the property value. It must
            accept a single argument which is the owner object. If not
            provided, the property cannot be deleted. The default is
            None.

        cached : bool, optional
            Whether or not the property caches the computed value. A
            cached property will only evaluate 'fget' once until the
            'reset' method of the property is invoked. The default is
            False.

        """
        gm = GetAttr.CachedProperty if cached else GetAttr.Property
        self.set_getattr_mode(gm, fget)
        self.set_setattr_mode(SetAttr.Property, fset)
        self.set_delattr_mode(DelAttr.Property, fdel)

    @property
    def fget(self):
        """ Get the getter function for the property.

        This will not find a specially named _get_* function.

        """
        return self.getattr_mode[1]

    @property
    def fset(self):
        """ Get the setter function for the property.

        This will not find a specially named _set_* function.

        """
        return self.setattr_mode[1]

    @property
    def fdel(self):
        """ Get the deleter function for the property.

        This will not find a specially named _del_* function.

        """
        return self.delattr_mode[1]

    @property
    def cached(self):
        """ Test whether or not this is a cached property.

        """
        return self.getattr_mode[0] == GetAttr.CachedProperty

    def getter(self, func):
        """ Use the given function as the property getter.

        This method is intended to be used as a decorator. The original
        function will still be callable.

        """
        mode, ignored = self.getattr_mode
        self.set_getattr_mode(mode, func)
        return func

    def setter(self, func):
        """ Use the given function as the property setter.

        This method is intended to be used as a decorator. The original
        function will still be callable.

        """
        self.set_setattr_mode(SetAttr.Property, func)
        return func

    def deleter(self, func):
        """ Use the given function as the property deleter.

        This method is intended to be used as a decorator. The original
        function will still be callable.

        """
        self.set_delattr_mode(DelAttr.Property, func)
        return func

    def reset(self, owner):
        """ Reset the value of the property.

        The old property value will be cleared and the notifiers will
        be run if the new value is different from the old value. If
        the property is not cached, notifiers will be unconditionally
        run using None as the old value.

        """
        reset_property(self, owner)


def cached_property(fget):
    """ A decorator which converts a function into a cached Property.

    Parameters
    ----------
    fget : callable
        The callable invoked to get the property value. It must accept
        a single argument which is the owner object.

    """
    return Property(fget, cached=True)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Validate
from .list import List


class ContainerList(List):
    """ A List member which supports container notifications.

    """
    __slots__ = ()

    def __init__(self, item=None, default=None):
        """ Initialize a ContainerList.

        """
        super(ContainerList, self).__init__(item, default)
        self.set_validate_mode(Validate.ContainerList, self.item)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .atom import AtomMeta, Atom, observe, set_default
from .catom import (
    CAtom, Member, GetAttr, SetAttr, PostGetAttr, PostSetAttr,
    DefaultValue, Validate, PostValidate, atomref, atomlist, atomclist
)
from .coerced import Coerced
from .containerlist import ContainerList
from .delegator import Delegator
from .dict import Dict
from .enum import Enum
from .event import Event
from .instance import Instance, ForwardInstance
from .intenum import IntEnum
from .list import List
from .property import Property, cached_property
from .scalars import (
    Value, ReadOnly, Constant, Callable, Bool, Int, Long, Range, Float, Bytes,
    Str, Unicode, FloatRange
)
from .signal import Signal
from .subclass import Subclass, ForwardSubclass
from .tuple import Tuple
from .typed import Typed, ForwardTyped



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
import sys
if sys.version_info >= (3,):
    import copyreg
else:
    import copy_reg as copyreg
from contextlib import contextmanager

from types import FunctionType
from future.utils import with_metaclass
from past.builtins import basestring

from .catom import (
    CAtom, Member, DefaultValue, PostGetAttr, PostSetAttr, Validate,
    PostValidate,
)


OBSERVE_PREFIX = '_observe_'
DEFAULT_PREFIX = '_default_'
VALIDATE_PREFIX = '_validate_'
POST_GETATTR_PREFIX = '_post_getattr_'
POST_SETATTR_PREFIX = '_post_setattr_'
POST_VALIDATE_PREFIX = '_post_validate_'


def observe(*names):
    """ A decorator which can be used to observe members on a class.

    Parameters
    ----------
    *names
        The str names of the attributes to observe on the object.
        These must be of the form 'foo' or 'foo.bar'.

    """
    # backwards compatibility for a single tuple or list argument
    if len(names) == 1 and isinstance(names[0], (tuple, list)):
        names = names[0]
    pairs = []
    for name in names:
        if not isinstance(name, basestring):
            msg = "observe attribute name must be a string, got '%s' instead"
            raise TypeError(msg % type(name).__name__)
        ndots = name.count('.')
        if ndots > 1:
            msg = "cannot observe '%s', only a single extension is allowed"
            raise TypeError(msg % name)
        if ndots == 1:
            name, attr = name.split('.')
            pairs.append((name, attr))
        else:
            pairs.append((name, None))
    return ObserveHandler(pairs)


class ObserveHandler(object):
    """ An object used to temporarily store observe decorator state.

    """
    __slots__ = ('pairs', 'func', 'funcname')

    def __init__(self, pairs):
        """ Initialize an ObserveHandler.

        Parameters
        ----------
        pairs : list
            The list of 2-tuples which store the pair information
            for the observers.

        """
        self.pairs = pairs
        self.func = None        # set by the __call__ method
        self.funcname = None    # storage for the metaclass

    def __call__(self, func):
        """ Called to decorate the function.

        """
        assert isinstance(func, FunctionType), "func must be a function"
        self.func = func
        return self

    def clone(self):
        """ Create a clone of the sentinel.

        """
        clone = type(self)(self.pairs)
        clone.func = self.func
        return clone


class set_default(object):
    """ An object used to set the default value of a base class member.

    """
    __slots__ = ('value', 'name')

    def __init__(self, value):
        self.value = value
        self.name = None    # storage for the metaclass

    def clone(self):
        """ Create a clone of the sentinel.

        """
        return type(self)(self.value)


class ExtendedObserver(object):
    """ A callable object used to implement extended observers.

    """
    __slots__ = ('funcname', 'attr')

    def __init__(self, funcname, attr):
        """ Initialize an ExtendedObserver.

        Parameters
        ----------
        funcname : str
            The function name on the owner object which should be
            used as the observer.

        attr : str
            The attribute name on the target object which should be
            observed.

        """
        self.funcname = funcname
        self.attr = attr

    def __call__(self, change):
        """ Handle a change of the target object.

        This handler will remove the old observer and attach a new
        observer to the target attribute. If the target object is not
        an Atom object, an exception will be raised.

        """
        old = None
        new = None
        ctype = change['type']
        if ctype == 'create':
            new = change['value']
        elif ctype == 'update':
            old = change['oldvalue']
            new = change['value']
        elif ctype == 'delete':
            old = change['value']
        attr = self.attr
        owner = change['object']
        handler = getattr(owner, self.funcname)
        if isinstance(old, Atom):
            old.unobserve(attr, handler)
        if isinstance(new, Atom):
            new.observe(attr, handler)
        elif new is not None:
            msg = "cannot attach observer '%s' to non-Atom %s"
            raise TypeError(msg % (attr, new))


class AtomMeta(type):
    """ The metaclass for classes derived from Atom.

    This metaclass computes the memory layout of the members in a given
    class so that the CAtom class can allocate exactly enough space for
    the object data slots when it instantiates an object.

    All classes deriving from Atom will be automatically slotted, which
    will prevent the creation of an instance dictionary and also the
    ability of an Atom to be weakly referenceable. If that behavior is
    required, then a subclasss should declare the appropriate slots.

    """
    def __new__(meta, name, bases, dct):
        # Unless the developer requests slots, they are automatically
        # turned off. This prevents the creation of instance dicts and
        # other space consuming features unless explicitly requested.
        if '__slots__' not in dct:
            dct['__slots__'] = ()

        # Pass over the class dict once and collect the information
        # necessary to implement the various behaviors. Some objects
        # declared on the class only serve as sentinels, and they are
        # removed from the dict before creating the class.
        observes = []               # Static observer methods: _observe_*
        defaults = []               # Default value methods: _default_*
        validates = []              # Validator methods: _validate_*
        decorated = []              # Decorated observers: @observe
        set_defaults = []           # set_default() sentinel
        post_getattrs = []          # Post getattr methods: _post_getattr_*
        post_setattrs = []          # Post setattr methods: _post_setattr_*
        post_validates = []         # Post validate methods: _post_validate_*
        seen_sentinels = set()      # The set of seen sentinels
        seen_decorated = set()      # The set of seen @observe decorators
        for key, value in dct.items():
            if isinstance(value, set_default):
                if value in seen_sentinels:
                    value = value.clone()
                value.name = key
                set_defaults.append(value)
                seen_sentinels.add(value)
                continue
            if isinstance(value, ObserveHandler):
                if value in seen_decorated:
                    value = value.clone()
                seen_decorated.add(value)
                decorated.append(value)
                value.funcname = key
                value = value.func
                dct[key] = value
                # Coninue processing the unwrapped function
            if isinstance(value, FunctionType):
                if key.startswith(OBSERVE_PREFIX):
                    observes.append(key)
                elif key.startswith(DEFAULT_PREFIX):
                    defaults.append(key)
                elif key.startswith(VALIDATE_PREFIX):
                    validates.append(key)
                elif key.startswith(POST_VALIDATE_PREFIX):
                    post_validates.append(key)
                elif key.startswith(POST_GETATTR_PREFIX):
                    post_getattrs.append(key)
                elif key.startswith(POST_SETATTR_PREFIX):
                    post_setattrs.append(key)

        # Remove the sentinels from the dict before creating the class.
        # The sentinels for the @observe decorators are already removed.
        for s in seen_sentinels:
            del dct[s.name]

        # Create the class object.
        cls = type.__new__(meta, name, bases, dct)

        # Walk the mro of the class, excluding itself, in reverse order
        # collecting all of the members into a single dict. The reverse
        # update preserves the mro of overridden members.
        members = {}
        for base in reversed(cls.__mro__[1:-1]):
            if base is not CAtom and issubclass(base, CAtom):
                members.update(base.__atom_members__)

        # The set of members which live on this class as opposed to a
        # base class. This enables the code which hooks up the various
        # static behaviors to only clone a member when necessary.
        owned_members = set()

        # Resolve any conflicts with memory layout. Conflicts can occur
        # with multiple inheritance where the indices of multiple base
        # classes will overlap. When this happens, the members which
        # conflict must be cloned in order to occupy a unique index.
        conflicts = []
        occupied = set()
        for member in members.values():
            if member.index in occupied:
                conflicts.append(member)
            else:
                occupied.add(member.index)

        # Clone the conflicting members and give them a unique index.
        # Do not blow away an overridden item on the current class.
        resolved_index = len(occupied)
        for member in conflicts:
            clone = member.clone()
            clone.set_index(resolved_index)
            owned_members.add(clone)
            members[clone.name] = clone
            if clone.name not in dct:
                setattr(cls, clone.name, clone)
            resolved_index += 1

        # Walk the dict a second time to collect the class members. This
        # assigns the name and the index to the member. If a member is
        # overriding an existing member, the memory index of the old
        # member is reused and any static observers are copied over.
        for key, value in dct.items():
            if isinstance(value, Member):
                if value in owned_members:  # foo = bar = Baz()
                    value = value.clone()
                    setattr(cls, key, value)
                owned_members.add(value)
                value.set_name(key)
                if key in members:
                    supermember = members[key]
                    members[key] = value
                    value.set_index(supermember.index)
                    value.copy_static_observers(supermember)
                else:
                    value.set_index(len(members))
                    members[key] = value

        # Add the special statically defined behaviors for the members.
        # If the target member is defined on a subclass, it is cloned
        # so that the behavior of the subclass is not modified.

        def clone_if_needed(m):
            if m not in owned_members:
                m = m.clone()
                members[m.name] = m
                owned_members.add(m)
                setattr(cls, m.name, m)
            return m

        # set_default() sentinels
        for sd in set_defaults:
            if sd.name not in members:
                msg = "Invalid call to set_default(). '%s' is not a member "
                msg += "on the '%s' class."
                raise TypeError(msg % (sd.name, name))
            member = clone_if_needed(members[sd.name])
            member.set_default_value_mode(DefaultValue.Static, sd.value)

        # _default_* methods
        n = len(DEFAULT_PREFIX)
        for mangled in defaults:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.set_default_value_mode(DefaultValue.ObjectMethod,
                                              mangled)

        # _validate_* methods
        n = len(VALIDATE_PREFIX)
        for mangled in validates:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.set_validate_mode(Validate.ObjectMethod_OldNew, mangled)

        # _post_validate_* methods
        n = len(POST_VALIDATE_PREFIX)
        for mangled in post_validates:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.set_post_validate_mode(PostValidate.ObjectMethod_OldNew,
                                              mangled)

        # _post_getattr_* methods
        n = len(POST_GETATTR_PREFIX)
        for mangled in post_getattrs:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.set_post_getattr_mode(PostGetAttr.ObjectMethod_Value,
                                             mangled)

        # _post_setattr_* methods
        n = len(POST_SETATTR_PREFIX)
        for mangled in post_setattrs:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.set_post_setattr_mode(PostSetAttr.ObjectMethod_OldNew,
                                             mangled)

        # _observe_* methods
        n = len(OBSERVE_PREFIX)
        for mangled in observes:
            target = mangled[n:]
            if target in members:
                member = clone_if_needed(members[target])
                member.add_static_observer(mangled)

        # @observe decorated methods
        for handler in decorated:
            for name, attr in handler.pairs:
                if name in members:
                    member = clone_if_needed(members[name])
                    observer = handler.funcname
                    if attr is not None:
                        observer = ExtendedObserver(observer, attr)
                    member.add_static_observer(observer)

        # Put a reference to the members dict on the class. This is used
        # by CAtom to query for the members and member count as needed.
        cls.__atom_members__ = members

        return cls


def __newobj__(cls, *args):
    """ A compatibility pickler function.

    This function is not part of the public Atom api.

    """
    return cls.__new__(cls, *args)


class Atom(with_metaclass(AtomMeta, CAtom)):
    """ The base class for defining atom objects.

    `Atom` objects are special Python objects which never allocate an
    instance dictionary unless one is explicitly requested. The storage
    for an atom is instead computed from the `Member` objects declared
    on the class. Memory is reserved for these members with no over
    allocation.

    This restriction make atom objects a bit less flexible than normal
    Python objects, but they are between 3x-10x more memory efficient
    than normal objects depending on the number of attributes.

    """

    @classmethod
    def members(cls):
        """ Get the members dictionary for the type.

        Returns
        -------
        result : dict
            The dictionary of members defined on the class. User code
            should not modify the contents of the dict.

        """
        return cls.__atom_members__

    @contextmanager
    def suppress_notifications(self):
        """ Disable member notifications within in a context.

        Returns
        -------
        result : contextmanager
            A context manager which disables atom notifications for the
            duration of the context. When the context exits, the state
            is restored to its previous value.

        """
        old = self.set_notifications_enabled(False)
        yield
        self.set_notifications_enabled(old)

    def __reduce_ex__(self, proto):
        """ An implementation of the reduce protocol.

        This method creates a reduction tuple for Atom instances. This
        method should not be overridden by subclasses unless the author
        fully understands the rammifications.

        """
        args = (type(self),) + self.__getnewargs__()
        return (__newobj__, args, self.__getstate__())

    def __getnewargs__(self):
        """ Get the argument tuple to pass to __new__ on unpickling.

        See the Python.org docs for more information.

        """
        return ()

    def __getstate__(self):
        """ The base implementation of the pickle getstate protocol.

        This base class implementation handles the generic case where
        the object and all of its state are pickable. This includes
        state stored in Atom members, as well as any instance dict or
        slot attributes. Subclasses which require further customization
        should reimplement this method and modify the dict generated by
        this base class method.

        """
        state = {}
        state.update(getattr(self, '__dict__', {}))
        slots = copyreg._slotnames(type(self))
        if slots:
            for name in slots:
                state[name] = getattr(self, name)
        for key in self.members():
            state[key] = getattr(self, key)
        return state

    def __setstate__(self, state):
        """ The base implementation of the pickle setstate protocol.

        This base class implementation handle the generic case of
        restoring an object using the state generated by the base
        class __getstate__ method. Subclasses which require custom
        behavior should reimplement this method.

        """
        for key, value in state.items():
            setattr(self, key, value)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
import sys
if sys.version_info >= (3,):
    from collections import MutableMapping
else:
    from UserDict import DictMixin as MutableMapping

from .catom import Member, PostGetAttr, DefaultValue, Validate
from .instance import Instance


class Dict(Member):
    """ A value of type `dict`.

    """
    __slots__ = ()

    def __init__(self, key=None, value=None, default=None):
        """ Initialize a Dict.

        Parameters
        ----------
        key : Member, type, tuple of types, or None, optional
            A member to use for validating the types of keys allowed in
            the dict. This can also be a type or a tuple of types, which
            will be wrapped with an Instance member. If this is not
            given, no key validation is performed.

        value : Member, type, tuple of types, or None, optional
            A member to use for validating the types of values allowed
            in the dict. This can also be a type or a tuple of types,
            which will be wrapped with an Instance member. If this is
            not given, no value validation is performed.

        default : dict, optional
            The default dict of items. A new copy of this dict will be
            created for each atom instance.

        """
        self.set_default_value_mode(DefaultValue.Dict, default)
        if key is not None and not isinstance(key, Member):
            key = Instance(key)
        if value is not None and not isinstance(value, Member):
            value = Instance(value)
        self.set_validate_mode(Validate.Dict, (key, value))
        if key is not None or value is not None:
            mode = PostGetAttr.MemberMethod_ObjectValue
            self.set_post_getattr_mode(mode, 'post_getattr')

    def post_getattr(self, owner, data):
        """ A post getattr handler.

        If the dict performs key or value validation, then this handler
        will be called to wrap the dict in a proxy object on the fly.

        """
        key, value = self.validate_mode[1]
        return _DictProxy(owner, key, value, data)

    def set_name(self, name):
        """ Assign the name to this member.

        This method is called by the Atom metaclass when a class is
        created. This makes sure the name of the internal members are
        also updated.

        """
        super(Dict, self).set_name(name)
        key, value = self.validate_mode[1]
        if key is not None:
            key.set_name(name + '|key')
        if value is not None:
            value.set_name(name + '|value')

    def set_index(self, index):
        """ Assign the index to this member.

        This method is called by the Atom metaclass when a class is
        created. This makes sure the index of the internal members are
        also updated.

        """
        super(Dict, self).set_index(index)
        key, value = self.validate_mode[1]
        if key is not None:
            key.set_index(index)
        if value is not None:
            value.set_index(index)


class _DictProxy(MutableMapping):
    """ A private proxy object which validates dict modifications.

    Instances of this class should not be created by user code.

    """
    # XXX move this class to C++
    def __init__(self, owner, keymember, valmember, data):
        self._owner = owner
        self._keymember = keymember
        self._valmember = valmember
        self._data = data

    def __repr__(self):
        return repr(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        owner = self._owner
        if self._keymember is not None:
            key = self._keymember.do_full_validate(owner, None, key)
        if self._valmember is not None:
            value = self._valmember.do_full_validate(owner, None, value)
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, key):
        return key in self._data

    def keys(self):
        return self._data.keys()

    def copy(self):
        return self._data.copy()

    def has_key(self, key):
        return key in self._data



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
# Note: This module is imported by 'atom.catom' module from code defined in
# the 'enumtypes.cpp' file. This module must therefore not import atom.
import sys
if sys.version_info >= (3,):
    import copyreg
else:
    import copy_reg as copyreg

from future.utils import with_metaclass
from past.builtins import basestring


# IntEnum is not defined until the metaclass creates it.
IntEnum = None


def _invalid_op(op):
    msg = " is an invalid operation for {}"
    msg = ("'%s'" % op) + msg

    def closure(self, *args):
        # Use format as otherwise fail for mod operator
        raise TypeError(msg.format(self))

    return closure


def _int_enum_flags_unpickler(enum_class, value):
    return int.__new__(enum_class.Flags, value)


def _int_enum_flags_pickler(flags):
    enum_class = type(flags).__enum_class__
    return _int_enum_flags_unpickler, (enum_class, int(flags))


class _IntEnumMeta(type):
    """ The metaclass for IntEnum and its subclasses.

    """
    class IntEnumFlags(int):

        # Set by the metaclass in the Flags property.
        __enum_class__ = None

        def __new__(cls, value=0):
            if isinstance(value, cls):
                return value
            if value == 0:
                return int.__new__(cls, value)
            return int.__new__(cls, cls.__enum_class__(value))

        def __make_flags(self, other, value):
            scls = type(self)
            ocls = type(other)
            if ocls is not scls and ocls is not scls.__enum_class__:
                msg = "cannot combine %s with object of type '%s'"
                raise TypeError(msg % (self, ocls.__name__))
            return int.__new__(scls, value)

        def __repr__(self):
            return '<enumflags: %s [value=%d]>' % (type(self).__name__, self)

        def __str__(self):
            return type(self).__name__

        def __and__(self, other):
            return self.__make_flags(other, int.__and__(self, other))

        __rand__ = __and__

        def __or__(self, other):
            return self.__make_flags(other, int.__or__(self, other))

        __ror__ = __or__

        def __xor__(self, other):
            return self.__make_flags(other, int.__xor__(self, other))

        __rxor__ = __xor__

        def __invert__(self):
            return self.__make_flags(self, int.__invert__(self))

        __add__ = _invalid_op('+')
        __sub__ = _invalid_op('-')
        __mul__ = _invalid_op('*')
        __div__ = _invalid_op('/')
        __truediv__ = _invalid_op('/')
        __floordiv__ = _invalid_op('//')
        __mod__ = _invalid_op('%')
        __divmod__ = _invalid_op('divmod')
        __pow__ = _invalid_op('**')
        __lshift__ = _invalid_op('<<')
        __rshift__ = _invalid_op('>>')
        __radd__ = _invalid_op('+')
        __rsub__ = _invalid_op('-')
        __rmul__ = _invalid_op('*')
        __rdiv__ = _invalid_op('/')
        __rtruediv__ = _invalid_op('/')
        __rfloordiv__ = _invalid_op('//')
        __rmod__ = _invalid_op('%')
        __rdivmod__ = _invalid_op('divmod')
        __rpow__ = _invalid_op('**')
        __rlshift__ = _invalid_op('<<')
        __rrshift__ = _invalid_op('>>')
        __neg__ = _invalid_op('-')
        __pos__ = _invalid_op('+')
        __abs__ = _invalid_op('abs')

    def __new__(meta, name, bases, dct):
        if len(bases) > 1:
            raise TypeError('int enums do not support multiple inheritance')
        if IntEnum is None:
            return type.__new__(meta, name, bases, dct)
        if bases[0] is not IntEnum:
            raise TypeError('int enums cannot be extended')
        enums = {}
        reved = {}
        cls = type.__new__(meta, name, bases, dct)
        for key, value in cls.__dict__.items():
            if isinstance(value, int):
                enum = int.__new__(cls, value)
                enum.__enum_name__ = key
                type.__setattr__(cls, key, enum)
                enums[key] = enum
                reved[int(enum)] = enum
        type.__setattr__(cls, '__enums__', enums)
        type.__setattr__(cls, '__reved__', reved)
        return cls

    def __call__(cls, which):
        enum = None
        if isinstance(which, int):
            enum = cls.__reved__.get(which)
        elif isinstance(which, basestring):
            enum = cls.__enums__.get(which)
        else:
            msg = "enum specifier must be an int or basestring, "
            msg += "got object of type '%s' instead"
            raise TypeError(msg % type(which).__name__)
        if enum is None:
            msg = "'%s' is not a valid %s enum specifier"
            raise ValueError(msg % (which, cls.__name__))
        return enum

    def __contains__(cls, which):
        return which in cls.__reved__ or which in cls.__enums__

    def __getitem__(cls, which):
        return cls(which)

    def __len__(cls):
        return len(cls.__enums__)

    def __iter__(cls):
        return iter(cls.__enums__.values())

    def __setattr__(cls, name, value):
        if name in cls.__enums__:
            raise TypeError("cannot change the value of an enum")
        type.__setattr__(cls, name, value)

    @property
    def Flags(cls):
        if cls.__flags_class__ is not None:
            return cls.__flags_class__
        name = cls.__name__ + 'Flags'
        flags_class = type(name, (cls.IntEnumFlags,), {})
        flags_class.__enum_class__ = cls
        cls.__flags_class__ = flags_class
        copyreg.pickle(flags_class, _int_enum_flags_pickler)
        return flags_class


class IntEnum(with_metaclass(_IntEnumMeta, int)):
    """ An integer subclass for declaring enum types.

    """

    # Set by the metaclass in the Flags property.
    __flags_class__ = None

    def __make_flags(self, other, value):
        scls = type(self)
        ocls = type(other)
        fcls = scls.Flags
        if ocls is not scls and ocls is not fcls:
            msg = "cannot combine %s with object of type '%s'"
            raise TypeError(msg % (self, ocls.__name__))
        return int.__new__(fcls, value)

    def __reduce_ex__(self, proto):
        return (type(self), (int(self),))

    def __repr__(self):
        t_name = type(self).__name__
        s_name = self.__enum_name__
        return '<enum: %s.%s [value=%d]>' % (t_name, s_name, self)

    def __str__(self):
        return '%s.%s' % (type(self).__name__, self.__enum_name__)

    def __and__(self, other):
        return self.__make_flags(other, int.__and__(self, other))

    __rand__ = __and__

    def __or__(self, other):
        return self.__make_flags(other, int.__or__(self, other))

    __ror__ = __or__

    def __xor__(self, other):
        return self.__make_flags(other, int.__xor__(self, other))

    __rxor__ = __xor__

    def __invert__(self):
        return self.__make_flags(self, int.__invert__(self))

    __add__ = _invalid_op('+')
    __sub__ = _invalid_op('-')
    __mul__ = _invalid_op('*')
    __div__ = _invalid_op('/')
    __truediv__ = _invalid_op('/')
    __floordiv__ = _invalid_op('//')
    __mod__ = _invalid_op('%')
    __divmod__ = _invalid_op('divmod')
    __pow__ = _invalid_op('**')
    __lshift__ = _invalid_op('<<')
    __rshift__ = _invalid_op('>>')
    __radd__ = _invalid_op('+')
    __rsub__ = _invalid_op('-')
    __rmul__ = _invalid_op('*')
    __rdiv__ = _invalid_op('/')
    __rtruediv__ = _invalid_op('/')
    __rfloordiv__ = _invalid_op('//')
    __rmod__ = _invalid_op('%')
    __rdivmod__ = _invalid_op('divmod')
    __rpow__ = _invalid_op('**')
    __rlshift__ = _invalid_op('<<')
    __rrshift__ = _invalid_op('>>')
    __neg__ = _invalid_op('-')
    __pos__ = _invalid_op('+')
    __abs__ = _invalid_op('abs')

    @property
    def name(self):
        return self.__enum_name__



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Member, DefaultValue, Validate


class Instance(Member):
    """ A value which allows objects of a given type or types.

    Values will be tested using the `PyObject_IsInstance` C API call.
    This call is equivalent to `isinstance(value, kind)` and all the
    same rules apply.

    The value of an Instance may be set to None.

    """
    __slots__ = ()

    def __init__(self, kind, args=None, kwargs=None, factory=None):
        """ Initialize an Instance.

        Parameters
        ----------
        kind : type or tuple of types
            The allowed type or types for the instance.

        args : tuple, optional
            If 'factory' is None, then 'kind' is a callable type and
            these arguments will be passed to the constructor to create
            the default value.

        kwargs : dict, optional
            If 'factory' is None, then 'kind' is a callable type and
            these keywords will be passed to the constructor to create
            the default value.

        factory : callable, optional
            An optional factory to use for creating the default value.
            If this is not provided and 'args' and 'kwargs' is None,
            then the default value will be None.

        """
        if factory is not None:
            self.set_default_value_mode(DefaultValue.CallObject, factory)
        elif args is not None or kwargs is not None:
            args = args or ()
            kwargs = kwargs or {}
            factory = lambda: kind(*args, **kwargs)
            self.set_default_value_mode(DefaultValue.CallObject, factory)
        self.set_validate_mode(Validate.Instance, kind)


class ForwardInstance(Instance):
    """ An Instance which delays resolving the type definition.

    The first time the value is accessed or modified, the type will
    be resolved and the forward instance will behave identically to
    a normal instance.

    """
    __slots__ = ('resolve', 'args', 'kwargs')

    def __init__(self, resolve, args=None, kwargs=None, factory=None):
        """ Initialize a ForwardInstance.

        resolve : callable
            A callable which takes no arguments and returns the type or
            tuple of types to use for validating the values.

        args : tuple, optional
            If 'factory' is None, then 'resolve' will return a callable
            type and these arguments will be passed to the constructor
            to create the default value.

        kwargs : dict, optional
            If 'factory' is None, then 'resolve' will return a callable
            type and these keywords will be passed to the constructor to
            create the default value.

        factory : callable, optional
            An optional factory to use for creating the default value.
            If this is not provided and 'args' and 'kwargs' is None,
            then the default value will be None.

        """
        self.resolve = resolve
        self.args = args
        self.kwargs = kwargs
        if factory is not None:
            self.set_default_value_mode(DefaultValue.CallObject, factory)
        elif args is not None or kwargs is not None:
            mode = DefaultValue.MemberMethod_Object
            self.set_default_value_mode(mode, "default")
        self.set_validate_mode(Validate.MemberMethod_ObjectOldNew, "validate")

    def default(self, owner):
        """ Called to retrieve the default value.

        This is called the first time the default value is retrieved
        for the member. It resolves the type and updates the internal
        default handler to behave like a normal Instance member.

        """
        kind = self.resolve()
        args = self.args or ()
        kwargs = self.kwargs or {}
        factory = lambda: kind(*args, **kwargs)
        self.set_default_value_mode(DefaultValue.CallObject, factory)
        return kind(*args, **kwargs)

    def validate(self, owner, old, new):
        """ Called to validate the value.

        This is called the first time a value is validated for the
        member. It resolves the type and updates the internal validate
        handler to behave like a normal Instance member.

        """
        kind = self.resolve()
        self.set_validate_mode(Validate.Instance, kind)
        return self.do_validate(owner, old, new)

    def clone(self):
        """ Create a clone of the ForwardInstance object.

        """
        clone = super(ForwardInstance, self).clone()
        clone.resolve = self.resolve
        clone.args = self.args
        clone.kwargs = self.kwargs
        return clone



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Member, DefaultValue, Validate
from .instance import Instance


class Tuple(Member):
    """ A member which allows tuple values.

    If item validation is used, then assignment will create a copy of
    the original tuple before validating the items, since validation
    may change the item values.

    """
    __slots__ = ()

    def __init__(self, item=None, default=()):
        """ Initialize a Tuple.

        Parameters
        ----------
        item : Member, type, or tuple of types, optional
            A member to use for validating the types of items allowed in
            the tuple. This can also be a type object or a tuple of types,
            in which case it will be wrapped with an Instance member. If
            this is not given, no item validation is performed.

        default : tuple, optional
            The default tuple of values.

        """
        if item is not None and not isinstance(item, Member):
            item = Instance(item)
        self.set_default_value_mode(DefaultValue.Static, default)
        self.set_validate_mode(Validate.Tuple, item)
